@@ -1,6 +1,6 @@
 /*
  * This file is part of ltrace.
- * Copyright (C) 2011,2012 Petr Machata, Red Hat Inc.
+ * Copyright (C) 2011,2012,2013 Petr Machata, Red Hat Inc.
  * Copyright (C) 1998,2004,2007,2008,2009 Juan Cespedes
  * Copyright (C) 2006 Ian Wienand
  * Copyright (C) 2006 Steve Fink
@@ -21,6 +21,8 @@
  * 02110-1301 USA
  */
 
+#define _XOPEN_SOURCE /* For wcwidth from wchar.h.  */
+
 #include <ctype.h>
 #include <stdlib.h>
 #include <assert.h>
@@ -28,13 +30,16 @@
 #include <stdarg.h>
 #include <stdio.h>
 #include <string.h>
+#include <wchar.h>
 
-#include "proc.h"
-#include "lens_default.h"
-#include "value.h"
+#include "config.h"
+#include "bits.h"
 #include "expr.h"
+#include "lens_default.h"
+#include "options.h"
+#include "output.h"
 #include "type.h"
-#include "common.h"
+#include "value.h"
 #include "zero.h"
 
 #define READER(NAME, TYPE)						\
@@ -56,9 +61,15 @@ READER(read_double, double)
 
 #undef READER
 
+#ifndef PRIb64  // Underlying format supported by bookworm glibc, but macros not exposed until 2.38
+#ifndef __PRI64_PREFIX
+#define __PRI64_PREFIX "ll"
+#endif
+#define PRIb64 __PRI64_PREFIX "b"
+#endif
 #define HANDLE_WIDTH(BITS)						\
 	do {								\
-		long l;							\
+		long long l;							\
 		if (value_extract_word(value, &l, arguments) < 0)	\
 			return -1;					\
 		int##BITS##_t i = l;					\
@@ -75,6 +86,8 @@ READER(read_double, double)
 			return fprintf(stream, "%"PRIu64, v);		\
 		case INT_FMT_o:						\
 			return fprintf(stream, "0%"PRIo64, v);		\
+		case INT_FMT_b:						\
+			return fprintf(stream, "%#"PRIb64, v);		\
 		}							\
 	} while (0)
 
@@ -84,6 +97,7 @@ enum int_fmt_t
 	INT_FMT_u,
 	INT_FMT_o,
 	INT_FMT_x,
+	INT_FMT_b,
 	INT_FMT_unknown,
 	INT_FMT_default,
 };
@@ -122,13 +136,8 @@ acc_fprintf(int *countp, FILE *stream, const char *format, ...)
 }
 
 static int
-format_char(FILE *stream, struct value *value, struct value_dict *arguments)
+print_char(FILE *stream, int c)
 {
-	long lc;
-	if (value_extract_word(value, &lc, arguments) < 0)
-		return -1;
-	int c = (int)lc;
-
 	const char *fmt;
 	switch (c) {
 	case -1:
@@ -172,13 +181,23 @@ format_char(FILE *stream, struct value *value, struct value_dict *arguments)
 }
 
 static int
-format_naked_char(FILE *stream, struct value *value,
-		  struct value_dict *arguments)
+format_char(FILE *stream, struct value *value, struct value_dict *arguments)
+{
+	long long lc;
+	if (value_extract_word(value, &lc, arguments) < 0)
+		return -1;
+	return print_char(stream, (int) lc);
+}
+
+static int
+format_naked(FILE *stream, struct value *value,
+	     struct value_dict *arguments,
+	     int (*what)(FILE *, struct value *, struct value_dict *))
 {
 	int written = 0;
 	if (acc_fprintf(&written, stream, "'") < 0
 	    || account_output(&written,
-			      format_char(stream, value, arguments)) < 0
+			      what(stream, value, arguments)) < 0
 	    || acc_fprintf(&written, stream, "'") < 0)
 		return -1;
 
@@ -231,7 +250,7 @@ static int
 format_struct(FILE *stream, struct value *value, struct value_dict *arguments)
 {
 	int written = 0;
-	if (acc_fprintf(&written, stream, "{ ") < 0)
+	if (acc_fprintf(&written, stream, "{") < 0)
 		return -1;
 
 	int need_delim = 0;
@@ -250,16 +269,17 @@ format_struct(FILE *stream, struct value *value, struct value_dict *arguments)
 
 		written += o;
 	}
-	if (acc_fprintf(&written, stream, " }") < 0)
+	if (acc_fprintf(&written, stream, "}") < 0)
 		return -1;
 	return written;
 }
 
+static const char null_message[] = "nil";
 int
 format_pointer(FILE *stream, struct value *value, struct value_dict *arguments)
 {
 	if (value_is_zero(value, arguments))
-		return fprintf(stream, "nil");
+		return fprintf(stream, null_message);
 
 	/* The following is for detecting recursion.  We keep track of
 	 * the values that were already displayed.  Each time a
@@ -321,7 +341,7 @@ format_pointer(FILE *stream, struct value *value, struct value_dict *arguments)
 	value_destroy(&element);
 
 done:
-	vect_popback(&pointers);
+	VECT_POPBACK(&pointers, struct value *, NULL, NULL);
 	return o;
 }
 
@@ -337,19 +357,27 @@ done:
  * OPEN, CLOSE, DELIM are opening and closing parenthesis and element
  *    delimiter.
  */
-int
+static int
 format_array(FILE *stream, struct value *value, struct value_dict *arguments,
 	     struct expr_node *length, size_t maxlen, int before,
 	     const char *open, const char *close, const char *delim)
 {
-	/* We need "long" to be long enough to cover the whole address
+	/* We need "long long" to be long enough to cover the whole address
 	 * space.  */
-	typedef char assert__long_enough_long[-(sizeof(long) < sizeof(void *))];
-	long l;
+	(void)sizeof(char[1 - 2*(sizeof(long long) < sizeof(void *))]);
+
+	int preloaded __attribute__((__cleanup__(value_preload_for_array_flush))) = -1;
+	if(!expr_is_trivial(length))  // zero()
+		preloaded = value_preload_for_array(value, maxlen + 1);
+
+	long long l;
 	if (expr_eval_word(length, value, arguments, &l) < 0)
 		return -1;
 	size_t len = (size_t)l;
 
+	if(preloaded == -1)
+		preloaded = value_preload_for_array(value, len > maxlen ? maxlen : len);
+
 	int written = 0;
 	if (acc_fprintf(&written, stream, "%s", open) < 0)
 		return -1;
@@ -394,18 +422,21 @@ toplevel_format_lens(struct lens *lens, FILE *stream,
 	case ARGTYPE_SHORT:
 	case ARGTYPE_INT:
 	case ARGTYPE_LONG:
+	case ARGTYPE_LLONG:
 		return format_integer(stream, value, int_fmt, arguments);
 
 	case ARGTYPE_USHORT:
 	case ARGTYPE_UINT:
 	case ARGTYPE_ULONG:
+	case ARGTYPE_ULLONG:
 		if (int_fmt == INT_FMT_i || int_fmt == INT_FMT_default)
 			int_fmt = INT_FMT_u;
 		return format_integer(stream, value, int_fmt, arguments);
 
 	case ARGTYPE_CHAR:
 		if (int_fmt == INT_FMT_default)
-			return format_naked_char(stream, value, arguments);
+			return format_naked(stream, value, arguments,
+					    &format_char);
 		return format_integer(stream, value, int_fmt, arguments);
 
 	case ARGTYPE_FLOAT:
@@ -416,14 +447,16 @@ toplevel_format_lens(struct lens *lens, FILE *stream,
 		return format_struct(stream, value, arguments);
 
 	case ARGTYPE_POINTER:
-		if (value->type->u.array_info.elt_type->type != ARGTYPE_VOID)
+		if (value_is_zero(value, arguments))
+			return fprintf(stream, null_message);
+		if (value->type->u.ptr_info.info->type != ARGTYPE_VOID)
 			return format_pointer(stream, value, arguments);
 		return format_integer(stream, value, INT_FMT_x, arguments);
 
 	case ARGTYPE_ARRAY:
 		return format_array(stream, value, arguments,
 				    value->type->u.array_info.length,
-				    options.arraylen, 1, "[ ", " ]", ", ");
+				    options.arraylen, 1, "[", "]", ", ");
 	}
 	abort();
 }
@@ -453,6 +486,18 @@ struct lens blind_lens = {
 };
 
 
+static int
+binary_lens_format_cb(struct lens *lens, FILE *stream,
+		     struct value *value, struct value_dict *arguments)
+{
+	return toplevel_format_lens(lens, stream, value, arguments, INT_FMT_b);
+}
+
+struct lens binary_lens = {
+	.format_cb = binary_lens_format_cb,
+};
+
+
 static int
 octal_lens_format_cb(struct lens *lens, FILE *stream,
 		     struct value *value, struct value_dict *arguments)
@@ -520,9 +565,11 @@ bool_lens_format_cb(struct lens *lens, FILE *stream,
 	case ARGTYPE_SHORT:
 	case ARGTYPE_INT:
 	case ARGTYPE_LONG:
+	case ARGTYPE_LLONG:
 	case ARGTYPE_USHORT:
 	case ARGTYPE_UINT:
 	case ARGTYPE_ULONG:
+	case ARGTYPE_ULLONG:
 	case ARGTYPE_CHAR:
 		if ((zero = value_is_zero(value, arguments)) < 0)
 			return -1;
@@ -538,6 +585,51 @@ struct lens bool_lens = {
 	.format_cb = bool_lens_format_cb,
 };
 
+static int
+redispatch_as_array(struct lens *lens, FILE *stream,
+		    struct value *value, struct value_dict *arguments,
+		    int (*cb)(struct lens *, FILE *,
+			      struct value *, struct value_dict *))
+{
+	struct arg_type_info info[2];
+	type_init_array(&info[1], value->type->u.ptr_info.info, 0,
+			expr_node_zero(), 0);
+	type_init_pointer(&info[0], &info[1], 0);
+	info->lens = lens;
+	info->own_lens = 0;
+	struct value tmp;
+	if (value_clone(&tmp, value) < 0)
+		return -1;
+	value_set_type(&tmp, info, 0);
+	int ret = cb(lens, stream, &tmp, arguments);
+	type_destroy(&info[0]);
+	type_destroy(&info[1]);
+	value_destroy(&tmp);
+	return ret;
+}
+
+static int
+format_wchar(FILE *stream, struct value *value, struct value_dict *arguments)
+{
+	long long l;
+	if (value_extract_word(value, &l, arguments) < 0)
+		return -1;
+	wchar_t wc = (wchar_t) l;
+	char buf[MB_CUR_MAX + 1];
+
+	int c = wctomb(buf, wc);
+	if (c < 0)
+		return -1;
+	if (c == 1)
+		return print_char(stream, buf[0]);
+
+	buf[c] = 0;
+	if (fprintf(stream, "%s", buf) < 0)
+		return -1;
+
+	c = wcwidth(wc);
+	return c >= 0 ? c : 0;
+}
 
 static int
 string_lens_format_cb(struct lens *lens, FILE *stream,
@@ -550,39 +642,43 @@ string_lens_format_cb(struct lens *lens, FILE *stream,
 		 * I suspect people are so used to the char * C idiom,
 		 * that string(char *) might actually turn up.  So
 		 * let's just support it.  */
-		if (value->type->u.ptr_info.info->type == ARGTYPE_CHAR) {
-			struct arg_type_info info[2];
-			type_init_array(&info[1],
-					value->type->u.ptr_info.info, 0,
-					expr_node_zero(), 0);
-			type_init_pointer(&info[0], &info[1], 0);
-			info->lens = lens;
-			info->own_lens = 0;
-			struct value tmp;
-			if (value_clone(&tmp, value) < 0)
-				return -1;
-			value_set_type(&tmp, info, 0);
-			int ret = string_lens_format_cb(lens, stream, &tmp,
-							arguments);
-			type_destroy(&info[0]);
-			type_destroy(&info[1]);
-			value_destroy(&tmp);
-			return ret;
-		}
+		switch ((int) value->type->u.ptr_info.info->type)
+		case ARGTYPE_CHAR:
+		case ARGTYPE_SHORT:
+		case ARGTYPE_USHORT:
+		case ARGTYPE_INT:
+		case ARGTYPE_UINT:
+		case ARGTYPE_LONG:
+		case ARGTYPE_LLONG:
+		case ARGTYPE_ULONG:
+		case ARGTYPE_ULLONG:
+			return redispatch_as_array(lens, stream, value,
+						   arguments,
+						   &string_lens_format_cb);
 
-		/* fall-through */
+		/* Otherwise dispatch to whatever the default for the
+		 * pointee is--most likely this will again be us.  */
+		/* Fall through.  */
 	case ARGTYPE_VOID:
 	case ARGTYPE_FLOAT:
 	case ARGTYPE_DOUBLE:
 	case ARGTYPE_STRUCT:
+		return toplevel_format_lens(lens, stream, value,
+					    arguments, INT_FMT_default);
+
 	case ARGTYPE_SHORT:
 	case ARGTYPE_INT:
 	case ARGTYPE_LONG:
+	case ARGTYPE_LLONG:
 	case ARGTYPE_USHORT:
 	case ARGTYPE_UINT:
 	case ARGTYPE_ULONG:
-		return toplevel_format_lens(lens, stream, value,
-					    arguments, INT_FMT_default);
+	case ARGTYPE_ULLONG:
+		if (value->parent != NULL && value->type->lens == NULL)
+			return format_wchar(stream, value, arguments);
+		else
+			return format_naked(stream, value, arguments,
+					    &format_wchar);
 
 	case ARGTYPE_CHAR:
 		return format_char(stream, value, arguments);
@@ -608,15 +704,6 @@ out_bits(FILE *stream, size_t low, size_t high)
 		return fprintf(stream, "%zd-%zd", low, high);
 }
 
-static unsigned
-bitcount(unsigned u)
-{
-	int c = 0;
-	for (; u > 0; u &= u - 1)
-		c++;
-	return c;
-}
-
 static int
 bitvect_lens_format_cb(struct lens *lens, FILE *stream,
 		       struct value *value, struct value_dict *arguments)
